import os
import pytest
import numpy as np
import easy_dna as dna


def test_extract_from_input(tmpdir):
    parts = []
    for i in range(10):
        part_id = "part_%s" % ("ABCDEFGHAB"[i])  # id is nonunique on purpose
        alias = "part_%d" % i  # alias is unique
        part_length = np.random.randint(1000, 1500)
        sequence = dna.random_dna_sequence(part_length)
        record = dna.sequence_to_biopython_record(sequence, id=part_id)
        record.name = part_id
        dna.annotate_record(record, label=part_id, alias=alias)
        parts.append(record)

    constructs = []
    for position_of_last_part in [8, 10]:
        # 8: parts A-H; 10: parts A--H and A, B again
        construct_record = sum(parts[1:position_of_last_part], parts[0])
        construct_record.id = "construct_%02d" % (position_of_last_part)
        construct_record.name = construct_record.id
        constructs.append(construct_record)

    target_dir = os.path.join(str(tmpdir), "test_dir")

    records_dict = dna.extract_from_input(
        construct_list=constructs, output_path=target_dir
    )
    assert records_dict["processed_report"]["shared_with"].count() == 16

    with pytest.raises(TypeError):
        dna.extract_from_input(output_path=target_dir)



import itertools
from .biotables import NUCLEOTIDE_TO_REGEXPR, IUPAC_NOTATION
from .biological_operations import reverse_complement
from .record_operations import sequence_to_atgc_string


def dna_pattern_to_regexpr(dna_pattern):
    """Return a regular expression pattern for the provided DNA pattern.

    For instance ``dna_pattern_to_regexpr('ATTNN')`` returns
    ``"ATT[A|T|G|C][A|T|G|C]"``.
    """
    return "".join([NUCLEOTIDE_TO_REGEXPR[n] for n in dna_pattern])


def all_iupac_variants(iupac_sequence):
    """Return all unambiguous possible versions of the given sequence.

    Examples
    --------
    >>> all_iupac_variants('ATN')
    ['ATA', 'ATC', 'ATG', 'ATT']
    """
    return [
        "".join(nucleotides)
        for nucleotides in itertools.product(
            *[IUPAC_NOTATION[n] for n in iupac_sequence]
        )
    ]


def find_index(seq, pattern):
    """Return the index of the first match of the pattern in seq.

    Returns -1 if there is no match.
    """

    pattern = sequence_to_atgc_string(pattern)
    if hasattr(seq, "seq"):
        # Seq is a SeqRecord
        return seq.seq.find(pattern)
    else:
        return seq.find(pattern)


def find_occurence(seq, pattern, strand="both"):
    if strand == 1:
        position = find_index(seq, pattern)
        if position == -1:
            return None
        else:
            return (position, position + len(pattern), 1)
    elif strand == -1:
        rev = reverse_complement(seq)
        occurrence = find_occurence(rev, pattern, strand=1)
        if occurrence is None:
            return None
        L = len(seq)
        return (L - occurrence[1], L - occurrence[0], -1)
    elif strand == "both":
        occurrence = find_occurence(seq, pattern, strand=1)
        if occurrence is None:
            occurrence = find_occurence(seq, pattern, strand=-1)
        return occurrence



from copy import deepcopy
from io import BytesIO, StringIO
import itertools
import os
import re

import pandas

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import crazydoc
import flametree
from snapgene_reader import snapgene_file_to_seqrecord

from .record_operations import sequence_to_biopython_record

strands = ["   ", "ss-", "ds-", "ms-"]
moleculetypes = [
    "NA    ",
    "DNA   ",
    "RNA   ",
    "tRNA  ",
    "rRNA  ",
    "mRNA  ",
    "uRNA  ",
    "cRNA  ",
]

strand_molecules = [
    "".join(entry).strip(" ") for entry in itertools.product(strands, moleculetypes)
]

genbank_reference = {
    # From the Genbank file format definition (release 260).
    # 0-based slicing indices below.
    # name: (start, end, [options])
    "locus": (0, 5, ["LOCUS"]),
    "name": (12, 28, None),
    "length": (29, 40, None),
    "bp": (41, 43, ["bp"]),
    "strandedness": (44, 47, strands),
    "moleculetype": (
        47,
        53,
        moleculetypes,
    ),
    # No topology info (spaces) also allowed, for flexibility:
    "moleculetopology": (55, 63, ["linear  ", "circular", "        "]),
    "divisioncode": (64, 67, None),
    "updatedate": (68, 79, None),
    # Spaces:
    "space1": (5, 12, "       "),
    "space2": (28, 29, " "),
    "space3": (40, 41, " "),
    "space4": (43, 44, " "),
    "space5": (53, 55, "  "),
    "space6": (63, 64, " "),
    "space7": (67, 68, " "),
}

genbank_months = [
    "JAN",
    "FEB",
    "MAR",
    "APR",
    "MAY",
    "JUN",
    "JUL",
    "AUG",
    "SEP",
    "OCT",
    "NOV",
    "DEC",
]

locus_guide = """+---+       +--------------+ +---------+ +- +--+----+  +------+ +-- +---------+
1   5       13            28 30       40 42 45 48  53  56    63 65  69       79"""


# Guidelines:
# Use only alphanumeric and underscore characters in the sequence name.
# Sequence names must be maximum 20 character long.
# Set SeqRecord.id to the sequence name, and .name="Exported" (this is used in exports).
# Genbank format must be: LOCUS = Exported, filename = sequence name.
# 'Exported' as name is within required length and also used by other software.
def load_record(
    filename, record_id="filename", adapt_id=True, upperize=False, id_cutoff=20
):
    """Load a FASTA/Genbank/Snapgene file as a Biopython record.

    Parameters
    ----------
    filename : str
        Path to the sequence file.
    record_id : str
        ID of the record ("filename": use the file name; "original": keep the record's
        original ID (defaults to file name if the record has no ID).
    adapt_id : bool
        If True, convert ID to alphanumeric and underscore.
    upperize : bool
        If True, the record's sequence will converted to uppercase.
    id_cutoff : int, optional
        If the ID is longer than this value, it will get truncated at this cutoff to
        conform to guidelines and Genbank name limit. Use `None` for no cutoff.
    """
    # FILEFORMAT
    if filename.lower().endswith((".gb", ".gbk")):
        record = SeqIO.read(filename, "genbank")
    elif filename.lower().endswith((".fa", ".fasta")):
        record = SeqIO.read(filename, "fasta")
    elif filename.lower().endswith(".dna"):
        record = snapgene_file_to_seqrecord(filename)
    else:
        raise ValueError("Unknown format for file: %s" % filename)
    # LETTERCASE
    if upperize:
        record = record.upper()
    # NAME
    record.name = "Exported"  # see guidelines above
    if record_id == "original":
        if record.id in [None, "", "<unknown id>", ".", " "]:
            record.id = os.path.splitext(os.path.basename(filename))[0]
    elif record_id == "filename":
        record.id = os.path.splitext(os.path.basename(filename))[0]
    else:  # to catch if a deprecated specification is used
        ValueError("`record_id` must be one of `filename` or `original`!")
    # LENGTH
    if id_cutoff is not None:
        record.id = record.id[:id_cutoff]
    # CHARACTERS
    if adapt_id:
        record.id = "".join([char if char.isalnum() else "_" for char in record.id])

    return record


def write_record(record, target, id_cutoff=20, adapt_id=True):
    """Write a DNA record as Genbank or FASTA via Biopython, with fixes.

    Parameters
    ----------
    record : SeqRecord
        Biopython SeqRecord.
    target : str or StringIO
        Filepath, string, or StringIO instance. Desired sequence format is inferred from
        the ending. If it's a directory, it uses the ID as filename and exports Genbank.
    id_cutoff : int, optional
        If the ID is longer than this value, it will get truncated at this cutoff to
        conform to guidelines and Genbank name limit. Use `None` for no cutoff.
    adapt_id : bool
        If True, convert ID to alphanumeric and underscore.
    """
    record = deepcopy(record)
    record.name = "Exported"  # This is used as LOCUS
    if id_cutoff is not None:
        record.id = record.id[:id_cutoff]
    if adapt_id:
        record.id = "".join([char if char.isalnum() else "_" for char in record.id])
    if has_dna_alphabet:  # for Biopython <1.78
        if str(record.seq.alphabet.__class__.__name__) != "DNAAlphabet":
            record.seq.alphabet = DNAAlphabet()
    record.annotations["molecule_type"] = "DNA"

    if hasattr(target, "open"):
        target = target.open("w")
    else:
        if type(target) is StringIO:
            fmt = "genbank"
        elif target.lower().endswith((".gb", ".gbk")):
            fmt = "genbank"
        elif target.lower().endswith((".fa", ".fasta")):
            fmt = "fasta"
        else:  # directory
            target = os.path.join(target, record.id + ".gb")
            fmt = "genbank"
    SeqIO.write(record, target, fmt)


def string_to_record(string):
    """Convert a string of a FASTA, Genbank... into a simple ATGC string.

    Can also be used to detect a format.
    """
    matches = re.match("([ATGC][ATGC]*)", string)
    # print("============", len(matches.groups()[0]), len(string))
    # print (matches.groups()[0] == string)
    if (matches is not None) and (matches.groups()[0] == string):
        if has_dna_alphabet:  # Biopython <1.78
            sequence = Seq(string, alphabet=DNAAlphabet())
        else:
            sequence = Seq(string)
        seqrecord = SeqRecord(sequence)
        seqrecord.annotations["molecule_type"] = "DNA"

        return seqrecord, "ATGC"

    for fmt in ("fasta", "genbank"):
        try:
            stringio = StringIO(string)
            records = list(SeqIO.parse(stringio, fmt))
            if len(records) > 0:
                return (records, fmt)
        except Exception:
            pass
    try:
        record = snapgene_file_to_seqrecord(filecontent=StringIO(string))
        return record
    except Exception:
        pass
    raise ValueError("Invalid sequence format")


def spreadsheet_file_to_dataframe(filepath, header="infer"):
    """Load a CSV or EXCEL spreadsheet as a Pandas dataframe."""
    name = filepath._name if hasattr(filepath, "_name") else filepath
    if name.endswith(".csv"):
        return pandas.read_csv(filepath, header=header)
    else:
        return pandas.read_excel(filepath, header=header)


def records_from_zip_file(zip_file):
    """Return all fasta/genbank/snapgene in a zip as Biopython records."""
    zip_file = flametree.file_tree(zip_file)
    records = []
    for f in zip_file._all_files:
        ext = f._extension.lower()
        if ext in ["gb", "gbk", "fa", "dna"]:
            try:
                new_records, fmt = string_to_record(f.read())
            except Exception:
                content_stream = BytesIO(f.read("rb"))
                try:
                    record = snapgene_file_to_seqrecord(fileobject=content_stream)
                    new_records, _ = [record], "snapgene"
                except Exception:
                    try:
                        parser = crazydoc.CrazydocParser(
                            ["highlight_color", "bold", "underline"]
                        )
                        new_records = parser.parse_doc_file(content_stream)
                        # fmt = "doc"
                    except Exception:
                        raise ValueError("Format not recognized for file " + f._path)

            single_record = len(new_records) == 1
            for i, record in enumerate(new_records):
                name = record.id
                if name in [
                    None,
                    "",
                    "<unknown id>",
                    ".",
                    " ",
                    "<unknown name>",
                ]:
                    number = "" if single_record else ("%04d" % i)
                    name = f._name_no_extension.replace(" ", "_") + number
                record.id = name
                record.name = name
                record.file_name = f._name_no_extension
            records += new_records
    return records


def records_from_file(filepath):
    """Autodetect file format and load Biopython records from it."""

    with open(filepath, "rb") as f:
        content = f.read()
    try:
        records, fmt = string_to_record(content.decode("utf-8"))
    except Exception:
        try:
            record = snapgene_file_to_seqrecord(fileobject=BytesIO(content))
            records, fmt = [record], "snapgene"
        except Exception:
            try:
                parser = crazydoc.CrazydocParser(
                    ["highlight_color", "bold", "underline"]
                )
                records = parser.parse_doc_file(BytesIO(content))
                fmt = "doc"
            except Exception:
                try:
                    df = spreadsheet_file_to_dataframe(filepath, header=None)
                    records = [
                        sequence_to_biopython_record(sequence=seq, id=name, name=name)
                        for name, seq in df.values
                    ]
                    fmt = "spreadsheet"
                except Exception:
                    raise ValueError("Format not recognized for file " + filepath)
    if not isinstance(records, list):
        records = [records]
    return records, fmt


def record_to_formated_string(record, remove_descr=False):
    """Return a string with the content of a Genbank file."""
    if remove_descr:
        record = deepcopy(record)
        if isinstance(record, (list, tuple)):
            for r in record:
                r.description = ""
        else:
            record.description = ""
    fileobject = StringIO()
    write_record(record=record, target=fileobject)

    return fileobject.getvalue().encode("utf-8")


def records_from_data_files(filepaths=None, folder=None):
    """Automatically convert files or a folder's content to Biopython records."""
    if folder is not None:
        filepaths = [f._path for f in flametree.file_tree(folder)._all_files]
    records = []
    for filepath in filepaths:
        filename = os.path.basename(filepath)
        if filename.lower().endswith("zip"):
            records += records_from_zip_file(filepath)
            continue
        recs, fmt = records_from_file(filepath)
        single_record = len(recs) == 1
        for i, record in enumerate(recs):
            name_no_extension = "".join(filename.split(".")[:-1])
            name = name_no_extension + ("" if single_record else ("%04d" % i))
            name = name.replace(" ", "_")
            UNKNOWN_IDS = [
                "None",
                "",
                "<unknown id>",
                ".",
                "EXPORTED",
                "<unknown name>",
                "Exported",
            ]
            if has_dna_alphabet:  # Biopython <1.78
                record.seq.alphabet = DNAAlphabet()
            record.annotations["molecule_type"] = "DNA"

            # Sorry for this parts, it took a lot of "whatever works".
            # keep your part names under 20c and pointless, and everything
            # will be good
            if str(record.id).strip() in UNKNOWN_IDS:
                record.id = name
            if str(record.name).strip() in UNKNOWN_IDS:
                record.name = name
            record.file_name = name_no_extension
        records += recs
    return records


def is_genbank_standard(filepath):
    """Check the LOCUS line of a Genbank file."""
    std_locus_line_len = 79  # as per Genbank definition
    is_correct = True
    expected_entries = 8  # expected, as per Genbank definition
    # Strandedness and Molecule Type is handled as one entry.

    with open(filepath) as genbank:
        first_line = genbank.readline().strip("\n")

    entries = re.split(" +", first_line)
    if len(entries) == expected_entries:
        topology = entries[-3]
    # Topology can be empty (spaces), in that case we expect one fewer column:
    elif len(entries) == expected_entries - 1:
        topology = "        "
    else:  # bad first line, quit
        print("Error: malformatted locus line: missing or too many entries")
        print(first_line, locus_guide, sep="\n")
        is_correct = False

        return is_correct

    entry_dict = {
        "locus": entries[0],
        "name": entries[1],
        "length": entries[2],
        "bp": entries[3],
        "strand_molecules": entries[4],
        "moleculetopology": topology,
        "divisioncode": entries[-2],
        "updatedate": entries[-1],
    }

    if len(first_line) == std_locus_line_len:
        # Look up each entry:
        for key, value in genbank_reference.items():
            if value[2] is not None:
                if first_line[value[0] : value[1]] not in value[2]:
                    print("Error:", key, "=", first_line[value[0] : value[1]])
                    print("       It should be one of ", value[2])
                    is_correct = False
    else:  # incorrect length, but we try and parse using whitespace
        print(
            "Error: LOCUS line is %d characters long! (standard length = %d)"
            % (
                len(first_line),
                std_locus_line_len,
            )
        )
        is_correct = False

        # This checks the same as the block above, except on whitespace-split entries:
        for entry, value in entry_dict.items():
            if entry in genbank_reference.keys():
                if genbank_reference[entry][2] is not None:
                    if value not in genbank_reference[entry][2]:
                        print("Error:", entry, "=", value)
                        print(
                            "       It should be one of ", genbank_reference[entry][2]
                        )
        # We split by whitespace, therefore this entry won't be in the Genbank ref
        if entry_dict["strand_molecules"] not in strand_molecules:
            print(
                "Error: incorrect strandness+molecule type = ",
                entry_dict["strand_molecules"],
            )

    # Inspecting entries in detail:
    if not entry_dict["length"].isnumeric():
        is_correct = False
        print("Error: non-numeric characters in length:", entries[2])
    # Date should be dd-MMM-yyyy (15-NOV-2024)
    date_values = entry_dict["updatedate"].split("-")
    if not date_values[0].isnumeric():  # ignores non-valid day number (e.g. 00)
        is_correct = False
        print("Error: non-numeric characters in day:", date_values[0])
    if date_values[1] not in genbank_months:
        is_correct = False
        print("Error: invalid month:", date_values[1])
    if not date_values[2].isnumeric():  # ignores non-valid year number
        is_correct = False
        print("Error: non-numeric characters in year:", date_values[2])

    if not is_correct:
        print(first_line, locus_guide, sep="\n")

    return is_correct



from .biological_operations import (
    complement,
    reverse_complement,
    translate,
    reverse_translate,
)

from .biotables import (
    CODONS_SEQUENCES,
    CODONS_SYNONYMS,
    CODONS_TRANSLATIONS,
    COMPLEMENTS,
)

from .io import load_record, write_record, records_from_data_files, is_genbank_standard

from .matching import dna_pattern_to_regexpr, all_iupac_variants

from .random_sequences import random_dna_sequence, random_protein_sequence

from .enzymes import list_common_enzymes

from .record_operations import (
    annotate_record,
    sequence_to_biopython_record,
    record_with_different_sequence,
    anonymized_record,
    censor_record,
    censor_genbank,
)

from .modifications import (
    replace_segment,
    replace_occurence,
    insert_segment,
    delete_segment,
    delete_nucleotides,
    reverse_segment,
    cut_and_paste_segment,
    copy_and_paste_segment,
    swap_segments,
)

from .extractor import extract_from_input

from .version import __version__

__all__ = [
    "complement",
    "reverse_complement",
    "translate",
    "reverse_translate",
    "CODONS_SEQUENCES",
    "CODONS_SYNONYMS",
    "CODONS_TRANSLATIONS",
    "COMPLEMENTS",
    "load_record",
    "write_record",
    "anonymized_record",
    "censor_record",
    "censor_genbank",
    "records_from_data_files",
    "is_genbank_standard",
    "dna_pattern_to_regexpr",
    "all_iupac_variants",
    "random_dna_sequence",
    "random_protein_sequence",
    "list_common_enzymes",
    "annotate_record",
    "sequence_to_biopython_record",
    "record_with_different_sequence",
    "replace_segment",
    "replace_occurence",
    "insert_segment",
    "delete_segment",
    "delete_nucleotides",
    "reverse_segment",
    "cut_and_paste_segment",
    "copy_and_paste_segment",
    "swap_segments",
    "extract_from_input",
]



from Bio.Seq import Seq
from easy_dna import (
    complement,
    reverse_complement,
    translate,
    reverse_translate,
    CODONS_TRANSLATIONS,
)


def test_complement():
    assert set(complement(Seq("A" * 31))) == set("T")
    assert set(complement("A" * 31)) == set("T")


def test_reverse_complement():
    assert reverse_complement(Seq("ATCG")) == str("CGAT")


def test_translate():
    assert translate("AAAAAC") == "KN"
    assert translate("AAAAAC", translation_table=CODONS_TRANSLATIONS) == "KN"


def test_reverse_translate():
    assert reverse_translate("KN") == "AAAAAC"



import pandas as pd
import flametree
from .io import load_record, records_from_data_files, write_record


def extract_from_input(
    filename=None,
    directory=None,
    construct_list=None,
    direct_sense=True,
    output_path=None,
    min_sequence_length=20,
):
    """Extract features from input and return in a dictionary.

    Optionally save the features in separate files.

    Parameters
    ----------
    filename : str
        Input sequence file path (e.g. Genbank).
    directory : str
        Directory name containing input sequence files.
    construct_list : list of SeqRecord
        A list of SeqRecords.
    direct_sense : bool
        If True, make antisense features into direct-sense in the exported files.
    output_path : str
        Path for the exported feature and report files.
    min_sequence_length : int
        Discard sequences with length less than this integer.
    """
    genbank_id_limit = 20  # GenBank format hard limit for name
    if construct_list:
        pass
    elif filename:
        input_record = load_record(
            filename, record_id="auto", upperize=False, id_cutoff=genbank_id_limit
        )
        construct_list = [input_record]
    elif directory:
        construct_list = records_from_data_files(filepaths=None, folder=directory)
    else:
        raise TypeError("Specify one of 'construct_list', 'filename' or 'directory'.")

    records_dict = dict()
    recordname_list = []
    for input_record in construct_list:
        records = extract_features(input_record, direct_sense)
        record_name = input_record.name[0:genbank_id_limit]
        # This part makes the key (used as dir name) unique by appending a copynumber:
        number_of_name_occurrences = recordname_list.count(record_name)
        if number_of_name_occurrences:
            key = "%s_%s" % (record_name, number_of_name_occurrences + 1)
        else:
            key = record_name

        recordname_list.append(record_name)

        records_dict[key] = records

    parts_report = make_part_dict(records_dict, min_sequence_length=min_sequence_length)
    processed_report = process_report(parts_report[1])

    all_parts_dict = parts_report[0]
    records_dict["all_parts"] = list(all_parts_dict.values())

    if output_path is not None:
        root = flametree.file_tree(output_path)

        for key, records in records_dict.items():

            record_dir = root._dir(key)

            record_name_alnum_list = []
            for record in records:

                record_name_alnum = "".join(
                    x if x.isalnum() else "_" for x in record.name
                )
                # This part makes the filename unique by appending a copynumber:
                number_of_occurrences = record_name_alnum_list.count(record_name_alnum)
                if number_of_occurrences:
                    record_filename = "%s_%s.gb" % (
                        record_name_alnum,
                        number_of_occurrences + 1,
                    )
                else:
                    record_filename = record_name_alnum + ".gb"

                record_name_alnum_list.append(record_name_alnum)

                record_file_path = record_dir._file(record_filename)

                try:
                    write_record(record, record_file_path, fmt="genbank")

                except Exception as err:
                    print("Error writing", record_filename, str(err))

        processed_report.to_csv(root._file("report.csv").open("w"))

    records_dict["processed_report"] = processed_report

    return records_dict


def compute_id(seq_feature):
    """Computes an ID for extract_features().

    The ID can be used as a SeqRecord/Genbank name or id attribute.
    """
    label_fields = [
        "name",
        "gene",
        "label",
        "product",
        "source",
        "note",
    ]

    label = "no_label"
    for key in label_fields:
        if key in seq_feature.qualifiers and len(seq_feature.qualifiers[key]):
            label = seq_feature.qualifiers[key]
            if isinstance(label, list):
                label = "_".join(label)
            break

    genbank_limit = 20  # GenBank format hard limit for name
    label = label[0:genbank_limit]
    label = label.replace(" ", "_").replace("'", "p")
    return label


def extract_features(seq_record, direct_sense=True):
    """Extract all features from a SeqRecord.

    Return the features as a list of SeqRecords.
    """
    all_features_in_direct_sense = direct_sense
    records = []

    # Check whether the feature is fully contained within any other feature:
    for i, seq_feature in enumerate(seq_record.features):
        start = seq_feature.location.start
        end = seq_feature.location.end

        contained_feature = False
        for other_seq in (
            seq_record.features[:i] + seq_record.features[(i + 1) :]
        ):  # exclude current feature
            if start in other_seq.location and end in other_seq.location:
                contained_feature = True
                break

        if contained_feature:
            continue

        if all_features_in_direct_sense:
            new_record = seq_feature.extract(seq_record)
        else:
            new_record = seq_record[
                seq_feature.location.start : seq_feature.location.end
            ]

        new_record.annotations = {"topology": "linear"}
        new_record.description = str(
            'Extracted from "' + new_record.id + '" by easy_dna.extract_features()'
        )

        new_record.id = compute_id(seq_feature)
        new_record.name = new_record.id
        records = records + [new_record]

    return records


def make_part_dict(records_dict, min_sequence_length=20):
    """Make a full part list and a report.

    Uses records_dict made by run_extraction().

    Parameters
    ----------
    records_dict : dict
        Dictionary of sequence name: list of features as SeqRecords.
    min_sequence_length : int
        Discard sequences with length less than this integer.
    """
    report_index = [
        "input_construct",
        "input_sequence",
        "shared_with",
        "equal_to",
        "has_copy",
        "too_short",
        "notes",
        "sequence_string",
    ]

    report = pd.DataFrame(
        data=None, index=report_index, columns=None, dtype=str, copy=False
    )
    all_parts_dict = dict()

    # This part is complex because it does two things, and could be simplified.
    # It makes a dictionary of all parts and a dataframe of part properties.
    # It also checks for a number of constraints and cases: sequence length,
    # shared sequences, shared names.
    for i, (key, record_list) in enumerate(records_dict.items()):
        for j, record in enumerate(record_list):
            s = pd.Series(None, index=report_index, name=i, dtype="float64")
            s["input_construct"] = key
            s["input_sequence"] = record.name
            s["has_copy"] = False  # default

            if len(record) < min_sequence_length:
                s["too_short"] = True
                series_name = str(i) + "_" + str(j)
                report[series_name] = s
                continue
            else:
                s["too_short"] = False

            sequence_as_key = str(record.seq.lower())
            s["sequence_string"] = sequence_as_key

            if sequence_as_key in all_parts_dict.keys():
                s["has_copy"] = True
            else:
                remove = 1
                while record.name in report.loc["input_sequence"].array:
                    index = len(record.name) - remove
                    record.name = record.name[:index] + "X" + record.name[index + 1 :]
                    remove += 1

                    s["note"] = "renamed"
                record.id = record.name

                s["input_sequence"] = record.name  # update name in record
                all_parts_dict[sequence_as_key] = record

            series_name = str(i) + "_" + str(j)
            report[series_name] = s

    report = report.T

    return (all_parts_dict, report)


def process_report(report):
    """Format the report prepared by make_part_dict().

    The function finds common parts within constructs and identical sequences
    between constructs.
    """
    all_shared_with = pd.Series(dtype="float64")
    all_equal_to = pd.Series(dtype="float64")
    for sequence in report.loc[report["has_copy"]]["sequence_string"]:
        constructs = report.loc[report["sequence_string"] == sequence][
            "input_construct"
        ]
        shared_with_text = "_|_".join(constructs.unique())
        shared_with = pd.Series(
            data=shared_with_text, index=constructs.index, dtype=str
        )

        all_shared_with = pd.concat([all_shared_with, shared_with])

        parts = report.loc[report["sequence_string"] == sequence]["input_sequence"]
        equal_to_text = "_|_".join(parts.unique())
        equal_to = pd.Series(data=equal_to_text, index=parts.index, dtype=str)

        all_equal_to = pd.concat([all_equal_to, equal_to])

    report.loc[all_equal_to.index, "equal_to"] = all_equal_to.values
    report.loc[all_shared_with.index, "shared_with"] = all_shared_with.values

    report.drop("has_copy", axis=1, inplace=True)

    return report



from .biological_operations import reverse_complement
from .matching import find_index


def replace_segment(seq, start, end, replacement):
    """Return the sequence with ``seq[start:end]`` replaced by ``replacement``."""
    return seq[:start] + replacement + seq[end:]


def insert_segment(seq, pos, inserted):
    """Return the sequence with ``inserted`` inserted, starting at index ``pos``."""
    return seq[:pos] + inserted + seq[pos:]


def delete_segment(seq, start, end):
    """Return the sequence with deleted segment from ``start`` to ``end``."""
    return seq[:start] + seq[end:]


def delete_nucleotides(seq, start, n):
    """Return the sequence with ``n`` deletions from position ``start``."""
    return seq[:start] + seq[start + n :]


def reverse_segment(seq, start, end):
    """Return the sequence with segment ``seq[start:end]`` reverse-complemented."""
    return seq[:start] + reverse_complement(seq[start:end]) + seq[end:]


def cut_and_paste_segment(seq, start, end, new_start):
    """Move a subsequence by "diff" nucleotides the left or the right."""
    sub = seq[start:end]
    diff = new_start - start
    if diff > 0:
        return seq[:start] + seq[end : end + diff] + sub + seq[end + diff :]
    else:
        return seq[: start + diff] + sub + seq[start + diff : start] + seq[end:]


def replace_occurence(seq, pattern, replacement, strand="both"):
    if strand == -1:
        rev_seq = replace_occurence(
            reverse_complement(seq), pattern, replacement, strand=1
        )
        return reverse_complement(rev_seq)
    elif strand == "both":
        new_seq = replace_occurence(seq, pattern, replacement, strand=1)
        if new_seq is seq:
            return replace_occurence(new_seq, pattern, replacement, strand=-1)
        else:
            return new_seq
    position = find_index(seq, pattern)
    if position == -1:
        return seq
    return replace_segment(seq, position, position + len(pattern), replacement)


def swap_segments(seq, pos1, pos2):
    """Return a new sequence with segments at position ``pos1`` and ``pos2`` swapped.

    ``pos1``, ``pos2`` are both of the form (start1, end1), (start2, end2).
    """
    (start1, end1), (start2, end2) = sorted([pos1, pos2])
    return (
        seq[:start1]
        + seq[start2:end2]
        + seq[end1:start2]
        + seq[start1:end1]
        + seq[end2:]
    )


def copy_and_paste_segment(seq, start, end, new_start):
    """Return the sequence with segment ``[start, end]`` also copied elsewhere,
    starting in ``new_start`."""
    return insert_segment(seq, new_start, seq[start:end])



from copy import deepcopy

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio.Restriction import Restriction

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False

from .random_sequences import random_dna_sequence
import easy_dna


# See io.py for an explanation of the approach regarding record operations.
def sequence_to_biopython_record(
    sequence, id="<unknown id>", name="Exported", features=()
):
    """Return a SeqRecord of a DNA sequence, ready to be Genbanked."""
    if hasattr(sequence, "seq"):  # already a SeqRecord()
        return sequence
    if has_dna_alphabet:  # Biopython <1.78
        sequence = Seq(sequence, alphabet=DNAAlphabet())
    else:
        sequence = Seq(sequence)

    seqrecord = SeqRecord(
        sequence,
        id=id,
        name=name,
        features=list(features),
    )
    seqrecord.annotations["molecule_type"] = "DNA"

    return seqrecord


def sequence_to_atgc_string(sequence):
    if isinstance(sequence, str):
        return sequence
    else:
        return str(sequence.seq)


def record_with_different_sequence(record, new_seq):
    """Return a version of the record with the sequence set to new_seq."""
    new_record = deepcopy(record)
    if has_dna_alphabet:  # Biopython <1.78
        sequence = Seq(new_seq, alphabet=DNAAlphabet())
    else:
        sequence = Seq(new_seq)

    new_record.seq = sequence

    return new_record


def annotate_record(
    seqrecord, location="full", feature_type="misc_feature", margin=0, **qualifiers
):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------
    seqrecord : SeqRecord
        The Biopython SeqRecord to be annotated.
    location : FeatureLocation
        Either (start, end) or (start, end, strand). (strand defaults to +1).
    feature_type : str
        The type associated with the feature.
    margin : int
        Number of extra bases added on each side of the given location.
    qualifiers : dict
        Dictionary that will be the Biopython feature's `qualifiers` attribute.
    """
    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(int(location[0]), int(location[1]), strand),
            qualifiers=qualifiers,
            type=feature_type,
        )
    )


def anonymized_record(record, record_id="anonymized", label_generator="feature_%d"):
    """Return a record with removed annotations/keywords/features/etc.

    Warning: this does not change the record sequence!

    Parameters
    ----------
    record : SeqRecord
        The record to be anonymized.
    record_id : str
        ID of the new record.
    label_generator : str or function
        Recipe to change feature labels. Either "feature_%d" or "None" (no label)
        or a function (i, feature) => label.
    """
    new_record = deepcopy(record)
    new_record.annotations = {
        "molecule_type": "ds-DNA",
        "data_file_division": "   ",
        "keywords": [],
    }
    new_record.id = record_id
    new_record.name = "Exported"  # in case it hasn't been set before
    for i, feature in enumerate(new_record.features):
        label = None
        if hasattr(label_generator, "__call__"):
            label = label_generator(i, feature)
        elif isinstance(label_generator, str):
            label = label_generator % i
        feature.qualifiers = {}
        if label is not None:
            feature.qualifiers["label"] = label
    return new_record


def censor_record(
    record,
    record_id="censored",
    label_generator="feature_%d",
    keep_topology=False,
    anonymise_features=True,
    preserve_sites=None,
):
    """Return a record with random sequence and censored annotations/features.

    Useful for creating example files or anonymizing sequences for bug reports.

    Parameters
    ----------
    record : SeqRecord
        The record to be anonymized.
    record_id : str
        ID of the new record.
    label_generator : str or function
        Recipe to change feature labels. Either "feature_%d" or "None" (no label) or a function (i, feature) => label.
    keep_topology : bool
        Whether to keep the record topology or not.
    anonymise_features : bool
        Whether to replace feature labels and ID/name, or not.
    preserve_sites : list of str
        List of enzyme sites to keep. Example: ["BsmBI", "BsaI"]. Preserves the sequence around cut sites of the specified enzymes.
    """
    # Anonymise
    if anonymise_features:
        new_record = anonymized_record(
            record, record_id=record_id, label_generator=label_generator
        )
    else:
        new_record = deepcopy(record)

    if keep_topology:
        try:
            new_record.annotations["topology"] = record.annotations["topology"]
        except KeyError:  # input may not have topology set
            pass

    # Randomise
    new_seq = random_dna_sequence(
        len(new_record), gc_share=None, probas=None, seed=None
    )

    if preserve_sites:
        restriction_batch = Restriction.RestrictionBatch(preserve_sites)
        # Destroy random new enzyme sites:
        analysis = Restriction.Analysis(restriction_batch, sequence=Seq(new_seq))
        analysis_results = analysis.full()
        for enzyme, hits in analysis_results.items():
            for hit in hits:
                # 10 bp up- and downstream destroys the site whichever strand it is on:
                if hit - 10 < 0:  # handle edge cases
                    start = 0
                    upstream = "A" * hit
                else:
                    start = hit - 10
                    upstream = "A" * 10
                if hit + 10 > len(new_seq):
                    end = len(new_seq)
                    downstream = "A" * (len(new_seq) - hit)
                else:
                    end = hit + 10
                    downstream = "A" * 10
                replacement = upstream + downstream
                new_seq = easy_dna.replace_segment(new_seq, start, end, replacement)

        # Add original sites:
        analysis = Restriction.Analysis(restriction_batch, sequence=record.seq)
        analysis_results = analysis.full()
        original_seq = str(record.seq)
        for enzyme, hits in analysis_results.items():
            for hit in hits:
                # keep 12 bp surrounding the cut site, to capture enzyme site:
                if hit - 12 < 0:  # handle edge cases
                    start = 0
                else:
                    start = hit - 12
                if hit + 12 > len(new_seq):
                    end = len(new_seq)
                else:
                    end = hit + 12
                original_segment = original_seq[start:end]
                new_seq = easy_dna.replace_segment(
                    new_seq, start, end, original_segment
                )

    censored_record = record_with_different_sequence(new_record, new_seq)

    return censored_record


def censor_genbank(filename, target, **censor_params):
    """Load Genbank file and write censored version.

    Parameters
    ----------
    filename : str
        Path to the file containing the record.
    target : str
        Path to output Genbank file.
    censor_params : dict, optional
        Optional parameters. See `censor_record()` for details.
    """
    record = easy_dna.load_record(filename)
    censored = easy_dna.censor_record(record, **censor_params)
    easy_dna.write_record(censored, target)



import os

from pandas import DataFrame

from Bio.SeqRecord import SeqRecord

from easy_dna import load_record, sequence_to_biopython_record
from easy_dna.io import (
    string_to_record,
    spreadsheet_file_to_dataframe,
    records_from_zip_file,
    records_from_file,
    record_to_formated_string,
    records_from_data_files,
    is_genbank_standard,
)

RECORDS_FOLDER = os.path.join("tests", "data")
CSV_FOLDER = os.path.join("easy_dna", "data")


def test_load_record():
    record = load_record(os.path.join(RECORDS_FOLDER, "test.gb"), upperize=True)
    assert isinstance(record, SeqRecord)


def test_string_to_record():
    records = string_to_record(">test\nATGC\n")
    assert len(records[0]) == 1  # list of records
    assert isinstance(records[0][0], SeqRecord)
    assert records[1] == "fasta"


def test_spreadsheet_file_to_dataframe():
    df = spreadsheet_file_to_dataframe(os.path.join(CSV_FOLDER, "complements.csv"))
    assert isinstance(df, DataFrame)


def test_records_from_zip_file():
    records = records_from_zip_file(os.path.join(RECORDS_FOLDER, "test.zip"))
    assert len(records) == 1
    assert isinstance(records[0], SeqRecord)


def test_records_from_file():
    record, fmt = records_from_file(os.path.join(RECORDS_FOLDER, "test.gb"))
    assert isinstance(record[0], SeqRecord)
    assert fmt == "genbank"


def test_record_to_formated_string():
    record = sequence_to_biopython_record(
        "ACGTGCGATGGGATTATTTCCAAC", id="test id", name="test name"
    )

    record_string = record_to_formated_string(record, remove_descr=True)
    assert isinstance(record_string, bytes)


def test_records_from_data_files():
    records = records_from_data_files(folder=RECORDS_FOLDER)
    for record in records:
        assert isinstance(record, SeqRecord)


def test_is_genbank_standard():
    assert is_genbank_standard(os.path.join(RECORDS_FOLDER, "test.gb")) == True



"""Methods to generate random sequences.

Also see utils.random_compatible_sequence() for random sequences verifying
user-provided constraints.
"""

import numpy as np


def random_dna_sequence(length, gc_share=None, probas=None, seed=None):
    """Return a random DNA sequence ("ATGGCGT...") with the specified length.

    Parameters
    ----------
    length : int
        Length of the DNA sequence.
    gc_share : float, optional
        The GC content of the random sequence, as a fraction (for example, 0.3 for 30%).
        Overwrites `probas`.
    probas : dict, optional
        Frequencies for the different nucleotides, for instance
        `probas={"A":0.2, "T":0.3, "G":0.3, "C":0.2}`.
        If not specified, all nucleotides are equiprobable (p=0.25).
    seed : int, optional
        The seed to feed to the random number generator. When a seed is provided,
        the random results depend deterministically on the seed, thus enabling
        reproducibility.
    """
    if seed is not None:
        np.random.seed(seed)
    if gc_share is not None:
        g_or_c = gc_share / 2.0
        not_g_or_c = (1 - gc_share) / 2.0
        probas = {"G": g_or_c, "C": g_or_c, "A": not_g_or_c, "T": not_g_or_c}
    if probas is None:
        sequence = np.random.choice(list("ATCG"), length)
    else:
        bases, probas = zip(*probas.items())
        sequence = np.random.choice(bases, length, p=probas)
    return "".join(sequence)


def random_protein_sequence(length, seed=None):
    """Return a random protein sequence "MNQTW...YL*" of the specified length.

    Parameters
    ----------
    length : int
        Length of the protein sequence (in number of amino acids). Note that the
        sequence will always start with "M" and end with a stop codon "*"
        with (length-2) random amino acids in the middle.
    seed : int, optional
        The seed to feed to the random number generator. When a seed is provided,
        the random results depend deterministically on the seed, thus enabling
        reproducibility.
    """
    if seed is not None:
        np.random.seed(seed)

    aa_list = list("ACEDGFIHKLNQPSRTWVY")
    aa_choices = np.random.choice(aa_list, length - 2)
    return "M" + "".join(aa_choices) + "*"



"""Methods for enzyme operations."""

from Bio import Restriction

from .matching import all_iupac_variants


def list_common_enzymes(
    site_length=(6,), opt_temp=(37,), min_suppliers=1, site_unlike=()
):
    """Return a list of enzyme names with the given constraints.

    Parameters
    ----------
    site_length : list of int
        List of accepted site lengths (6, 4, ...).
    opt_temp : list of int
        List of accepted optimal temperatures for the enzyme.
    min_suppliers : int
        Minimal number of registered suppliers in the Biopython data. A minimum
        of 3 known suppliers returns the most common enzymes.
    site_unlike : list of str
        List of (ambiguous or unambiguous) DNA sequences that should NOT be
        recognized by the selected enzymes.
    """
    site_unlike = set(
        [
            variant
            for enzyme in site_unlike
            for variant in all_iupac_variants(Restriction.__dict__[enzyme].site)
        ]
    )

    def is_valid(enzyme_name):
        enzyme = Restriction.__dict__[enzyme_name]
        return (
            len(enzyme.site) in site_length
            and enzyme.opt_temp in opt_temp
            and len(enzyme.supplier_list()) >= min_suppliers
            and len(set(all_iupac_variants(enzyme.site)).intersection(site_unlike)) == 0
        )

    return [
        enzyme_name
        for enzyme_name in Restriction.AllEnzymes.elements()
        if is_valid(enzyme_name)
    ]



import os

from easy_dna import (
    annotate_record,
    sequence_to_biopython_record,
    anonymized_record,
    censor_record,
    censor_genbank,
)

RECORDS_FOLDER = os.path.join("tests", "data")


def test_anonymized_record():
    record = sequence_to_biopython_record(
        "ACGTGCGATGGGATTATTTCCAAC", id="test id", name="test name"
    )
    annotate_record(record, location="full", feature_type="test_feature")
    record = anonymized_record(record)
    assert record.seq == "ACGTGCGATGGGATTATTTCCAAC"
    assert record.id == "anonymized"
    assert record.name == "Exported"
    assert record.features[0].qualifiers["label"] == "feature_0"


def test_censor_record():
    sequence = "ACGTGCGAT" + "CGTCTC" + "ATGC" + "A" * 20
    record = sequence_to_biopython_record(sequence, id="test id", name="test name")
    annotate_record(record, location="full", feature_type="test_feature")
    record = censor_record(
        record, keep_topology=True, anonymise_features=True, preserve_sites=["BsmBI"]
    )
    assert record.seq != sequence
    assert record.id == "censored"
    assert record.name == "Exported"
    assert record.features[0].qualifiers["label"] == "feature_0"
    assert "CGTCTC" in record.seq


def test_censor_genbank(tmpdir):
    filename = os.path.join(RECORDS_FOLDER, "test.gb")
    target = os.path.join(tmpdir, "censored.gb")
    censor_genbank(filename, target)
    assert os.path.isfile(target)



from easy_dna import (
    dna_pattern_to_regexpr,
    record_with_different_sequence,
    sequence_to_biopython_record,
    annotate_record,
    list_common_enzymes,
)


def test_record_with_different_sequence():
    record = sequence_to_biopython_record("ATGCATGCATGC")
    annotate_record(record, (0, 5), label="my_label")
    new_record = record_with_different_sequence(record, "GGCCGGCCGGCCGGCC")
    assert len(new_record.features) == 1
    assert new_record.features[0].location == record.features[0].location


def test_list_common_enzymes():
    assert len(list_common_enzymes(min_suppliers=3)) >= 61
    # biopython v1.80: 61 enzymes
    # < v1.79: 63 enzymes



from easy_dna import random_dna_sequence, random_protein_sequence


def test_random_dna_sequence():
    n = 15
    assert len(random_dna_sequence(n, seed=42)) == n

    random_dna_no_GC = random_dna_sequence(n, gc_share=0, seed=42)
    assert len(random_dna_no_GC) == n
    assert len(set("AT") | set(random_dna_no_GC)) == len(set("AT"))


def test_random_protein_sequence():
    n = 10
    random_protein = random_protein_sequence(n, seed=2)
    assert random_protein[0] == "M"
    assert random_protein[-1] == "*"
    assert len(random_protein) == n
    aa_set = set("ACEDGFIHKLNQPSRTWVY" + "M*")
    assert len(aa_set | set(random_protein)) == len(aa_set)



import numpy as np

from Bio.Seq import Seq

from .biotables import COMPLEMENTS, CODONS_SEQUENCES


def complement(dna_sequence):
    """Return the complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"TACGGC"``.

    Uses Biopython for speed.
    """
    if hasattr(dna_sequence, "complement"):
        return dna_sequence.complement()
    if len(dna_sequence) <= 30:
        return "".join([COMPLEMENTS[nuc] for nuc in dna_sequence])
    # This alternative has overhead but is really fast on long sequences
    return str(Seq(dna_sequence).complement())


def reverse_complement(dna_sequence):
    """Return the reverse-complement of the DNA sequence.

    For instance ``reverse_complement("ATGCCG")`` returns ``"CGGCAT"``.

    Uses Biopython for speed.
    """
    if hasattr(dna_sequence, "reverse_complement"):
        return dna_sequence.reverse_complement()
    return complement(dna_sequence)[::-1]


def reverse_translate(protein_sequence, randomize_codons=False):
    """Return a DNA sequence which translates to the provided protein sequence.

    Note: at the moment, the first valid codon found is used for each
    amino-acid (so it is deterministic but no codon-optimization is done).
    """
    if randomize_codons:
        random_indices = np.random.randint(0, 1000, len(protein_sequence))
        return "".join(
            [
                CODONS_SEQUENCES[aa][random_index % len(CODONS_SEQUENCES[aa])]
                for aa, random_index in zip(protein_sequence, random_indices)
            ]
        )
    return "".join([CODONS_SEQUENCES[aa][0] for aa in protein_sequence])


def translate(dna_sequence, translation_table="Bacterial"):
    """Translate the DNA sequence into an amino-acid sequence "MLKYQT...".

    If ``translation_table`` is the name or number of a NCBI genetic table,
    Biopython will be used. See here for options:

    http://biopython.org/DIST/docs/tutorial/Tutorial.html#sec26

    ``translation_table`` can also be a dictionary of the form
    ``{"ATT": "M", "CTC": "X", etc.}`` for more exotic translation tables.
    """
    if isinstance(translation_table, dict):
        return "".join(
            [
                translation_table[dna_sequence[i : i + 3]]
                for i in range(0, len(dna_sequence), 3)
            ]
        )
    else:
        return str(Seq(dna_sequence).translate(table=translation_table))



from easy_dna import (
    replace_segment,
    replace_occurence,
    insert_segment,
    delete_segment,
    delete_nucleotides,
    reverse_segment,
    cut_and_paste_segment,
    copy_and_paste_segment,
    swap_segments,
)


def test_replace_segment():
    assert replace_segment("ATGCATGCTACGGGCT", 2, 5, "AAAAAA") == "ATAAAAAATGCTACGGGCT"


def test_insert_segment():
    assert insert_segment("ATGCGGGCT", 2, "AAAAAA") == "ATAAAAAAGCGGGCT"


def test_delete_segment():
    assert delete_segment("ATGCATGCTACGGGCT", 2, 5) == "ATTGCTACGGGCT"


def test_delete_nucleotides():
    assert delete_nucleotides("AAAAAAGCGGGCT", 0, 6) == "GCGGGCT"


def test_reverse_segment():
    assert reverse_segment("AAAAGTTCCAAAA", 4, 9) == "AAAAGGAACAAAA"


def test_cut_and_paste_segment():
    assert cut_and_paste_segment("AAAAGTTCCAAAA", 4, 9, 0) == "GTTCCAAAAAAAA"
    assert cut_and_paste_segment("AAAAGTTCCAAAA", 4, 9, 13) == "AAAAAAAAGTTCC"


def test_replace_occurence():
    assert replace_occurence("AAAGTTCCA", "GGAA", "CCCC", strand="both") == "AAAGGGGGA"


def test_swap_segments():
    assert swap_segments("AAAGTTCCA", pos1=(0, 3), pos2=(8, 9)) == "AGTTCCAAA"


def test_copy_and_paste_segment():
    assert copy_and_paste_segment("AAAGTTCCA", 0, 3, 9) == "AAAGTTCCAAAA"



"""This module provides useful biological tables as Python dictionaries.
"""

from collections import defaultdict
import os


def reverse_table(table):
    """Return a dictionary {v1: [k1a, k1b,...]} where k1a, k1b are all
    the keys of table such that table[k1]=v1.
    """
    new_table = defaultdict(lambda: [])
    for (k, v) in sorted(table.items()):
        new_table[v].append(k)
    return dict(new_table)


def dict_from_csv(filepath, sep=";"):
    """Read a CSV and store entries in a dict.

    Warning: function is used internally. Does not work properly if quoted
    cells contain separators, for example: `"ent;ry1";entry2`.
    """
    with open(filepath, "r") as f:
        return {
            line.split(sep)[0]: line.split(sep)[1].strip("\n")
            for line in f.readlines()
            if line not in ("", "\n")
        }


data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")

# TABLE DEFINITIONS START HERE

# COMPLEMENTS = {"A": "T", "T": "A", "C": "G", "G": "C"}
with open(os.path.join(data_dir, "complements.csv"), "r") as f:
    COMPLEMENTS = dict([line.split(",") for line in f.read().splitlines()])
OTHER_BASES = {
    "A": ["T", "G", "C"],
    "T": ["A", "G", "C"],
    "G": ["T", "A", "C"],
    "C": ["T", "G", "A"],
}

CODONS_TRANSLATIONS = dict_from_csv(os.path.join(data_dir, "codons_translations.csv"))
CODONS_SEQUENCES = reverse_table(CODONS_TRANSLATIONS)
CODONS_SYNONYMS = {
    codon: CODONS_SEQUENCES[CODONS_TRANSLATIONS[codon]] for codon in CODONS_TRANSLATIONS
}

NUCLEOTIDE_TO_REGEXPR = dict_from_csv(
    os.path.join(data_dir, "nucleotide_to_regexpr.csv")
)


iupac_file = os.path.join(data_dir, "iupac_notation.csv")
IUPAC_NOTATION = {k: set(v) for k, v in dict_from_csv(iupac_file).items()}



from easy_dna import dna_pattern_to_regexpr, all_iupac_variants
from easy_dna.matching import find_index, find_occurence


def test_dna_pattern_to_regexpr():
    assert dna_pattern_to_regexpr("ATW") == "AT[ATW]"


def test_all_iupac_variants():
    assert sorted(all_iupac_variants("ATN")) == ["ATA", "ATC", "ATG", "ATT"]


def test_find_index():
    assert find_index("ATCGATTCGATC", "CGA") == 2


def test_find_occurence():
    assert find_occurence('ATCGATTCGATC', 'CGA', strand="both") == (2, 5, 1)
    assert find_occurence('ATCGATTCGATC', 'AAAAA', strand="both") == None
    assert find_occurence('ATCGATTCGATC', 'AATC', strand="both") == (3, 7, -1)
